@@ -30,6 +30,7 @@
 #define _GNU_SOURCE
 #include "internal.h"
 #include "backend_utils.h"
+#include "linux_desktop_settings.h"
 
 #include <X11/Xresource.h>
 
@@ -614,10 +615,12 @@ Cursor _glfwCreateCursorX11(const GLFWimage* image, int xhot, int yhot)
 //////                       GLFW platform API                      //////
 //////////////////////////////////////////////////////////////////////////
 
-GLFWAPI int glfwGetCurrentSystemColorTheme(void) {
-    return 0;
+GLFWAPI GLFWColorScheme glfwGetCurrentSystemColorTheme(bool query_if_unintialized) {
+    return glfw_current_system_color_theme(query_if_unintialized);
 }
 
+void _glfwPlatformInputColorScheme(GLFWColorScheme appearance UNUSED) { }
+
 int _glfwPlatformInit(void)
 {
     XInitThreads();
@@ -646,11 +649,13 @@ int _glfwPlatformInit(void)
                         "X11: Failed to initialize event loop data");
     }
     glfw_dbus_init(&_glfw.x11.dbus, &_glfw.x11.eventLoopData);
+    glfw_initialize_desktop_settings();  // needed for color scheme change notification
 
     _glfw.x11.screen = DefaultScreen(_glfw.x11.display);
     _glfw.x11.root = RootWindow(_glfw.x11.display, _glfw.x11.screen);
     _glfw.x11.context = XUniqueContext();
     _glfw.x11.RESOURCE_MANAGER = XInternAtom(_glfw.x11.display, "RESOURCE_MANAGER", True);
+    _glfw.x11._KDE_NET_WM_BLUR_BEHIND_REGION = None;
     XSelectInput(_glfw.x11.display, _glfw.x11.root, PropertyChangeMask);
 
     _glfwGetSystemContentScaleX11(&_glfw.x11.contentScaleX, &_glfw.x11.contentScaleY, false);
